__description__ = "a module that tests several aspects of the io submodule"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#--------------------------------------------------------------------------

import copy
import sys
import pytest

import numpy as np

from idq import io
from idq import utils

#--------------------------------------------------------------------------

#----------------------
#        tests
#----------------------

@pytest.mark.usefixtures('gstlal_classifierdata')
class TestGstlalHDF5ClassifierData(object):
    """
    Tests several aspects of GstlalClassifierData for a compliant dataset.
    """
    def test_times(self, gstlal_classifierdata_conf):
        assert self.classifierdata.start == gstlal_classifierdata_conf['start_time'], 'incorrect start time set'
        assert self.classifierdata.end == gstlal_classifierdata_conf['end_time'], 'incorrect end time set'
        
    def test_segs(self, gstlal_classifierdata_conf):
        assert self.classifierdata.segs == gstlal_classifierdata_conf['segs'], 'incorrect segments returned for ClassifierData'

    def test_columns(self, gstlal_classifierdata_conf):
        self.classifierdata.triggers(channels=gstlal_classifierdata_conf['channels'])
        assert set(self.classifierdata.columns) == set(gstlal_classifierdata_conf['columns']), 'incorrect columns retrieved from ClassifierData'

    def test_channels(self, gstlal_classifierdata_conf):
        self.classifierdata.triggers(channels=gstlal_classifierdata_conf['channels'])
        assert set(self.classifierdata.channels) == set(gstlal_classifierdata_conf['channels']), 'incorrect channels stored in ClassifierData'

    def test_trigger_query_simple(self, gstlal_classifierdata_conf):
        result = self.classifierdata.triggers()
        for channel in gstlal_classifierdata_conf['channels']:
            assert channel in self.classifierdata.channels, 'channel not present in ClassifierData'

    def test_trigger_query_channels(self, gstlal_classifierdata_conf):
        channels2query = gstlal_classifierdata_conf['channels'][1:]
        result = self.classifierdata.triggers(channels=channels2query)
        for channel in channels2query:
            assert channel in result.keys(), 'channel not present in ClassifierData that was requested'

    def test_trigger_query_columns(self, gstlal_classifierdata_conf):
        columns2query = ['time', 'snr']
        result = self.classifierdata.triggers(columns=columns2query)
        for channel in self.classifierdata.channels:
            for column in columns2query:
                try:
                    result[channel][column]
                except KeyError:
                    raise KeyError('column not present in ClassifierData that was requested')

    def test_trigger_query_segments(self, gstlal_classifierdata_conf):
        segs2query = copy.copy(gstlal_classifierdata_conf['segs'])
        segs2query.contract(100)
        segs_remainder = gstlal_classifierdata_conf['segs'] - segs2query
        result = self.classifierdata.triggers(segs=segs2query, time='time')
        for channel in self.classifierdata.channels:
            for seg in segs_remainder:
                 result[channel] = utils.filter_triggers(result[channel], bounds={'time': (float(seg[0]), float(seg[1]))})
            assert not result[channel], 'segment query result from ClassifierData contains data outside of bounds'

    def test_pop(self, gstlal_classifierdata_conf):
        self.classifierdata.triggers(channels=gstlal_classifierdata_conf['channels'])
        test_channel = gstlal_classifierdata_conf['channels'][0]
        test_triggers = self.classifierdata.pop(test_channel)
        assert test_channel not in self.classifierdata.channels, 'popped channel still present in ClassifierData'

        triggers = self.classifierdata.triggers(channels=gstlal_classifierdata_conf['channels'])
        assert triggers[test_channel].shape == test_triggers.shape, 'triggers from popped channel and triggers contained in ClassifierData do not have the same shape'
        assert bool(np.asarray(test_triggers == triggers[test_channel]).all()), 'triggers from popped channel do not match triggers contained in ClassifierData'
